#include <cstdarg>
#include <cstdlib>
#include <cstdio>
#include "shallowwater.h"

using namespace std;

void ShallowWater::
init(float lx_, int cell_nx, int cell_nz)
{
   nx=cell_nx;
   nz=cell_nz;
   h.init(nx, nz);
   u.init(nx+1, nz);  // staggered grid for velocity
   w.init(nx, nz+1);  // staggered grid for velocity
   solidh.init(nx, nz);
   solidu.init(nx+1, nz);  // staggered grid for velocity
   solidw.init(nx, nz+1);  // staggered grid for velocity
   newh.init(nx, nz);
   newu.init(nx+1, nz);  // staggered grid for velocity
   neww.init(nx, nz+1);  // staggered grid for velocity
   lx=lx_;
   lz=(lx*nz)/nx;
   dx=lx/nx;
   overdx=nx/lx;
   density=1000;
   gravity=9.81;
}

void ShallowWater::
read_config(const char *filename)
{
   int i, j;
   FILE *in=fopen(filename, "rt");
   if(!in){
      fprintf(stderr, "couldn't read config file %s\n", filename);
      return;
   }

   // first read in dimensions of grid
   int cell_nx, cell_nz;
   float lx_;
   fscanf(in, "%d %d", &cell_nx, &cell_nz);
   fscanf(in, "%f", &lx_);
   init(lx_, cell_nx, cell_nz);

   // next read in solid level set
   // for simplicity this is sampled at twice the grid cell resolution (so it includes the staggered points too)
   for(j=0; j<2*cell_nz+1; ++j) for(i=0; i<2*cell_nx+1; ++i){
      float phi;
      fscanf(in, "%f", &phi);
      if(i%2 && j%2){ // if this is a cell centre
	 solidh((i-1)/2, (j-1)/2)=phi;
      }else if(j%2){ // if instead it's a face centre for u
         solidu(i/2, (j-1)/2)=phi;
      }else if(i%2){ // if instead it's a face centre for w
         solidw((i-1)/2, j/2)=phi;
      }
   }

   // read in the initial heights
   for(j=0; j<cell_nz; ++j) for(i=0; i<cell_nx; ++i){
      fscanf(in, "%f", &h(i,j));
   }

   // read in the initial u field
   for(j=0; j<cell_nz; ++j) for(i=0; i<cell_nx+1; ++i){
      fscanf(in, "%f", &u(i,j));
   }

   // read in the initial w field
   for(j=0; j<cell_nz+1; ++j) for(i=0; i<cell_nx; ++i){
      fscanf(in, "%f", &w(i,j));
   }
   
   // and clean up
   fclose(in);
}

void ShallowWater::
write(const char *filename_format, ...)
{
   va_list ap;
   va_start(ap, filename_format);
#ifdef _MSC_VER
#define FILENAMELENGTH 256
   char filename[FILENAMELENGTH];
   vsnprintf(filename, FILENAMELENGTH, filename_format, ap);
   FILE *out=fopen(filename, "wt");
#else 
   char *filename;
   vasprintf(&filename, filename_format, ap);
   FILE *out=fopen(filename, "wt");
   free(filename);
#endif
   va_end(ap);

   // we'll actually write something in the same format as the config file, to make it easy to do restarts
   int i, j;
   // write grid dimensions
   fprintf(out, "%d %d\n%g\n", h.nx, h.ny, lx);
   // write solid level set data, at twice grid resolution
   for(j=0; j<2*h.ny+1; ++j){
      for(i=0; i<2*h.nx+1; ++i){
         float phi;
         if(i%2 && j%2){ // if this is a cell centre
            phi=solidh((i-1)/2, (j-1)/2);
         }else if(j%2){ // if instead it's a face centre for u
            phi=solidu(i/2, (j-1)/2);
         }else if(i%2){ // if instead it's a face centre for w
            phi=solidw((i-1)/2, j/2);
         }else{
	    phi=0; // default value for unneeded data
	 }
         fprintf(out, " %g", phi);
      }
      fprintf(out, "\n");
   }
   fprintf(out, "\n");
   // write heights
   for(j=0; j<h.ny; ++j){
      for(i=0; i<h.nx; ++i)
         fprintf(out, " %g", h(i,j));
      fprintf(out, "\n");
   }
   fprintf(out, "\n");
   // write u
   for(j=0; j<u.ny; ++j){
      for(i=0; i<u.nx; ++i)
         fprintf(out, " %g", u(i,j));
      fprintf(out, "\n");
   }
   fprintf(out, "\n");
   // write w
   for(j=0; j<w.ny; ++j){
      for(i=0; i<w.nx; ++i)
         fprintf(out, " %g", w(i,j));
      fprintf(out, "\n");
   }
   // and clean up
   fclose(out);
}

float ShallowWater::
maxdt(void)
{
   //@@@
   //@@@ Calculate a maximum stable time step. You may or may not want to limit this based
   //@@@ on advection (the CFL condition), but you will need to take into account the wave
   //@@@ speed.
}

void ShallowWater::
advance(float dt)
{
   // we assume the solid arrays and the solid arrays have been set, and the
   // old velocity vectors inside the solids are correct (we don't require new velocity
   // vectors inside the new solid positions)
   fill_in_old_height();     // fill in the gaps where the solids are
   advect_height(dt);        // move height around non-conservatively
   fix_height(dt);           // change height depending on divergence of velocity
   advect_velocity(dt);      // move velocity around non-conservatively
   apply_pressure(dt);       // add in gravity term
   switch_old_and_new();
}

void ShallowWater::
fill_in_old_height(void)
{
   int i, j;
   // do one sweep to propagate heights from lower i or lower j
   for(j=0; j<nz; ++j) for(i=0; i<nx; ++i) if(solidh(i,j)<0){
      float w=0, wh=0; // used to compute weighted average of heights
      if(i>0    && solidh(i-1,j)>=solidh(i,j)){ w+=1; wh+=h(i-1,j); }
      if(i<nx-1 && solidh(i+1,j)>=solidh(i,j)){ w+=1; wh+=h(i+1,j); }
      if(j>0    && solidh(i,j-1)>=solidh(i,j)){ w+=1; wh+=h(i,j-1); }
      if(j<nz-1 && solidh(i,j+1)>=solidh(i,j)){ w+=1; wh+=h(i,j+1); }
      if(w>0) h(i,j)=wh/w; // use average if any neighbours were found
   }
   // sweep in the opposite directions to fill in the rest
   for(j=nz-1; j>=0; --j) for(i=nx-1; i>=0; --i) if(solidh(i,j)<0){
      float w=0, wh=0; // used to compute weighted average of heights
      if(i>0    && solidh(i-1,j)>=solidh(i,j)){ w+=1; wh+=h(i-1,j); }
      if(i<nx-1 && solidh(i+1,j)>=solidh(i,j)){ w+=1; wh+=h(i+1,j); }
      if(j>0    && solidh(i,j-1)>=solidh(i,j)){ w+=1; wh+=h(i,j-1); }
      if(j<nz-1 && solidh(i,j+1)>=solidh(i,j)){ w+=1; wh+=h(i,j+1); }
      if(w>0) h(i,j)=wh/w; // use average if any neighbours were found
   }
}

static float interp1d(float qneg1, float q0, float q1, float q2, float x)
{
   //@@@ Implement the minmod-limited cubic Hermite interpolation in 1 dimension
   //@@@ The q data is assumed to be given at grid points -1, 0, 1, and 2, and
   //@@@ the parameter x is somewhere between 0 and 1. Evaluate the cubic at x and
   //@@@ return the value.
}

static float interpolate(const Array2f &a, float x, float y)
{
   int i, ilo, ihi, j, jlo, jhi;
   float fx, fy;

   if(x<0){ i=0; fx=0; }
   else if(x>=a.nx-1){ i=a.nx-2; fx=1; }
   else{ i=(int)x; fx=x-i; }
   if(i==0) ilo=0; else ilo=i-1;
   if(i==a.nx-2) ihi=i+1; else ihi=i+2;

   if(y<0){ j=0; fy=0; }
   else if(y>=a.ny-1){ j=a.ny-2; fy=1; }
   else{ j=(int)y; fy=y-j; }
   if(j==0) jlo=0; else jlo=j-1;
   if(j==a.ny-2) jhi=j+1; else jhi=j+2;

   // do a tensor-product of 1d interpolations
   return interp1d(interp1d(a(ilo,jlo), a(i,jlo), a(i+1,jlo), a(ihi,jlo), fx),
                   interp1d(a(ilo,j), a(i,j), a(i+1,j), a(ihi,j), fx),
                   interp1d(a(ilo,j+1), a(i,j+1), a(i+1,j+1), a(ihi,j+1), fx),
                   interp1d(a(ilo,jhi), a(i,jhi), a(i+1,jhi), a(ihi,jhi), fx), fy);
}

void ShallowWater::
advect_height(float dt)
{
   int i, j;
   for(j=0; j<nz; ++j) for(i=0; i<nx; ++i) if(solidh(i,j)>=0){
      //@@@ write the semi-Lagrangian advection code for height
      //@@@ That is, figure out what oldx and oldz should be, based on
      //@@@ the position (i,j) of the grid point and the grid-scaled
      //@@@ velocity u*overdx, w*overdx.
      newh(i,j)=interpolate(h, oldx, oldz);
   }
}

void ShallowWater::
fix_height(float dt)
{
   int i, j;
   for(j=0; j<nz; ++j) for(i=0; i<nx; ++i) if(solidh(i,j)>=0){
      //@@@ write the forcing term on the height based on divergence of velocity
      //@@@ That is, update newh(i,j) from u and w values.

      // in addition, make sure the height doesn't erroneously go negative
      if(newh(i,j)<=0)
	 newh(i,j)=0;
   }
}

void ShallowWater::
advect_velocity(float dt)
{
   //@@@ figure out the semi-Lagrangian advection for both velocity components,
   //@@@ getting newu and neww in terms of u and w.
}

void ShallowWater::
apply_pressure(float dt)
{
   //@@@ adjust newu and neww according to gravity and the gradient of the height field.
}

void ShallowWater::
switch_old_and_new(void)
{
   // now that we have computed the new values of h, u, and w, copy them to the old values
   // (actually implemented by switching array pointers)
   float *temp;
   temp=h.data; h.data=newh.data; newh.data=temp;
   temp=u.data; u.data=newu.data; newu.data=temp;
   temp=w.data; w.data=neww.data; neww.data=temp;
}

